@@ -29,6 +29,7 @@ import uuid
 from keystoneauth1 import loading
 from oslo_config import cfg
 from oslo_log import log as logging
+from oslo_utils import timeutils
 import requests
 import six
 
@@ -43,10 +44,19 @@ from castellan.i18n import _
 from castellan.key_manager import key_manager
 
 DEFAULT_VAULT_URL = "http://127.0.0.1:8200"
+DEFAULT_MOUNTPOINT = "secret"
 
 vault_opts = [
     cfg.StrOpt('root_token_id',
                help='root token for vault'),
+    cfg.StrOpt('approle_role_id',
+               help='AppRole role_id for authentication with vault'),
+    cfg.StrOpt('approle_secret_id',
+               help='AppRole secret_id for authentication with vault'),
+    cfg.StrOpt('kv_mountpoint',
+               default=DEFAULT_MOUNTPOINT,
+               help='Mountpoint of KV store in Vault to use, for example: '
+                    '{}'.format(DEFAULT_MOUNTPOINT)),
     cfg.StrOpt('vault_url',
                default=DEFAULT_VAULT_URL,
                help='Use this endpoint to connect to Vault, for example: '
@@ -88,6 +98,12 @@ class VaultKeyManager(key_manager.KeyManager):
         loading.register_session_conf_options(self._conf, VAULT_OPT_GROUP)
         self._session = requests.Session()
         self._root_token_id = self._conf.vault.root_token_id
+        self._approle_role_id = self._conf.vault.approle_role_id
+        self._approle_secret_id = self._conf.vault.approle_secret_id
+        self._cached_approle_token_id = None
+        self._approle_token_ttl = None
+        self._approle_token_issue = None
+        self._kv_mountpoint = self._conf.vault.kv_mountpoint
         self._vault_url = self._conf.vault.vault_url
         if self._vault_url.startswith("https://"):
             self._verify_server = self._conf.vault.ssl_ca_crt_file or True
@@ -104,7 +120,10 @@ class VaultKeyManager(key_manager.KeyManager):
         if self._vault_kv_version:
             return self._vault_kv_version
 
-        resource_url = self._get_url() + 'v1/sys/internal/ui/mounts/secret'
+        resource_url = '{}v1/sys/internal/ui/mounts/{}'.format(
+            self._get_url(),
+            self._kv_mountpoint
+        )
         resp = self._do_http_request(self._session.get, resource_url)
 
         if resp.status_code == requests.codes['not_found']:
@@ -115,8 +134,9 @@ class VaultKeyManager(key_manager.KeyManager):
         return self._vault_kv_version
 
     def _get_resource_url(self, key_id=None):
-        return '{}v1/secret/{}{}'.format(
+        return '{}v1/{}/{}{}'.format(
             self._get_url(),
+            self._kv_mountpoint,
 
             '' if self._get_api_version() == '1' else
             'data/' if key_id else
@@ -124,9 +144,58 @@ class VaultKeyManager(key_manager.KeyManager):
 
             key_id if key_id else '?list=true')
 
+    @property
+    def _approle_token_id(self):
+        if (all((self._approle_token_issue, self._approle_token_ttl)) and
+                timeutils.is_older_than(self._approle_token_issue,
+                                        self._approle_token_ttl)):
+            self._cached_approle_token_id = None
+        return self._cached_approle_token_id
+
+    def _build_auth_headers(self):
+        if self._root_token_id:
+            return {'X-Vault-Token': self._root_token_id}
+
+        if self._approle_token_id:
+            return {'X-Vault-Token': self._approle_token_id}
+
+        if self._approle_role_id:
+            params = {
+                'role_id': self._approle_role_id
+            }
+            if self._approle_secret_id:
+                params['secret_id'] = self._approle_secret_id
+            approle_login_url = '{}v1/auth/approle/login'.format(
+                self._get_url()
+            )
+            token_issue_utc = timeutils.utcnow()
+            try:
+                resp = self._session.post(url=approle_login_url,
+                                          json=params,
+                                          verify=self._verify_server)
+            except requests.exceptions.Timeout as ex:
+                raise exception.KeyManagerError(six.text_type(ex))
+            except requests.exceptions.ConnectionError as ex:
+                raise exception.KeyManagerError(six.text_type(ex))
+            except Exception as ex:
+                raise exception.KeyManagerError(six.text_type(ex))
+
+            if resp.status_code in _EXCEPTIONS_BY_CODE:
+                raise exception.KeyManagerError(resp.reason)
+            if resp.status_code == requests.codes['forbidden']:
+                raise exception.Forbidden()
+
+            resp = resp.json()
+            self._cached_approle_token_id = resp['auth']['client_token']
+            self._approle_token_issue = token_issue_utc
+            self._approle_token_ttl = resp['auth']['lease_duration']
+            return {'X-Vault-Token': self._approle_token_id}
+
+        return {}
+
     def _do_http_request(self, method, resource, json=None):
         verify = self._verify_server
-        headers = {'X-Vault-Token': self._root_token_id}
+        headers = self._build_auth_headers()
 
         try:
             resp = method(resource, headers=headers, json=json, verify=verify)
